//
//  Project   : besy
//  File      : bank.cc
//  Author    : Ronald Kriemann / Christian Rost
//  Purpose   : class for a bankserver
//  Date      : 25.05.1997
//  Changes   : 
//

#include <stdio.h>


#include <iostream.h>

#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/msg.h>

#include "bank.hh"
#include "../misc/base64.hh"
#include "../misc/TString.hh"
#include "../misc/msqdef.hh"

#ifdef SUNOS
#  define STREAMIO
#endif

//
// constructor
// - gets msq
//
Bank::Bank ()
{     
	key_t  key = MSQKEY;					
	// Create Bankdata
	KontoListe = new TList<TKonto>();
	KundenListe = new TList<TKunde>();
	       
#ifdef STREAMIO
	//Load old Bankdata
	if (load("bankdata.dat"))       
	    printf("Successfully loaded  bankdata.dat\n");	    	  
	else
	{ 
	    printf("Failed to load bankdata.dat.\nI try to create new bankdata and save it to file bankdata.dat\n");
	    initBankdata(); 
	    printf("Savesuccess: %i \n", save("bankdata.dat"));
	}
#else
	initBankdata();
#endif
		
	// create msgqueue
	if ((msqid = msgget(key, IPC_CREAT | MSQPERM)) < 0)
		perror( "Bank::Bank () : msgget" );  	
}

//
// Einmaliges Erstellen der Konten und Kunden 
//
void 
Bank::initBankdata()
{  
	// Initiale Kunden-Konten-Generierung		 	                                      
	createKunde(14,                                              //Kundennummer
				"besy14",                                        //Name
				"CY9rzUYh03PK3k6DJie09g==",                      //md5(test)     Passwort
				"krDazX4Sad6+o5kuJCCkgg==",                      //100*md5(test) Onetimepasswort
				"Pe4dolH1RD7GDmPKIgJKgEUAs3S/kBvlF7ihm9wpZrs="); //RSA-Modulus vom Kunden
	createKonto(14,getKunde(14),20000);	  

  //besy-Konten erstellen  
	createKunde(1,
				"besy1",
				encrypt("besy1"),
				"Q1eRF2+Ytyx+pI1FOC0i9w",
				"");
	createKonto(1,getKunde(1),1000);

	for (int i=10;i<=20;i++)
	{
		char passwd[10];	     
		sprintf(passwd,"besy%d",i);
		switch (i)
		{
            case 14: break;
            case 13: createKunde(i,passwd,encrypt(passwd),"kl9gSovuEX4k4iesZGK/Vg==", ""); break;
            case 16: createKunde(i,passwd,encrypt(passwd),"1B2M2Y8AsgTpgAmY7PhCfg==", ""); break;
            default: createKunde(i,passwd,encrypt(passwd),"onepw", ""); break;
		}// if
		createKonto(i,getKunde(i),1000);	     
	}// for
		
	// Save the new data to file
	

}


//
// destructor
// - remove msq and delete bank-data
//
Bank::~Bank ()
{
#ifdef DEBUG
	cout << "Bank::~Bank () : deleting msq" << endl;
#endif	
	if (msgctl( msqid, IPC_RMID, (struct msqid_ds*) 0) < 0)
		perror( "Bank::~Bank () : msgctl" );

	// Delete Bank
	delete KontoListe;
	delete KundenListe;
}

TString
Bank::encrypt( const TString &input )
{
	return TBase64::encode( md5.crypt(input) );
}

bool 
Bank::createKonto(int Nummer, TKunde *Kunde, double Betrag=0)
{	
	if (Kunde==0) return FALSE;
	TKonto Konto(Nummer,Kunde->getNummer());
	if (Betrag !=0)
    {
		TTransaktion *Aktion = new TTransaktion(0,Nummer,Betrag,2,"1. Einzahlung");
		Konto.addAktion(*Aktion);
    }// if  
	// Check ob Kontonummer doppelt vorhanden
	if ( ! KontoListe->isIn(Konto) )
    {
		KontoListe->addElement(Konto);
		return TRUE;
    }// if  
	return FALSE;     
}


//Fuegt einen Kunden zur Kundenliste dazu
bool 
Bank::createKunde ( int nummer, 
					const TString &Name, 
					const TString &Pass,
					const TString &OnePass, 
					const TString &key)
{  
	TKunde Kunde(nummer,Name,Pass,OnePass, key);  

	// Check ob Kunde vorhanden
	if ( ! KundenListe->isIn(Kunde))
	{
		KundenListe->addElement(Kunde);
		return TRUE;
	}// if  

	return FALSE;     
}



bool 
Bank::addTransaktion ( int Source,int Target,double Betrag,
					   const TString &Zweck,int Modus)
{  
	switch (Modus) 
	{
        case  DauerAuftrag :    
        case Ueber : 
            if (Source == 0) return FALSE; 
            if (Betrag <=0 ) return FALSE; 
            if (Source == Target) return FALSE; 
            break; // Syntaxfehler

        case BarEin : 
            if ( Betrag <=0) return FALSE;
            break;

        case BarAus : 
            if (Source !=0) return FALSE; 
            if (Betrag <=0) return FALSE; 
            break; //Syntaxfehler

        default: 
            return FALSE;
	}// switch


	TTransaktion *Aktion = new TTransaktion(Source,Target,Betrag,Modus,Zweck);
	TKonto *SourceKonto=0;
	TKonto *TargetKonto=0;  

	if (Modus==4) 
	{
		// Checken ob SourceKonto existiert
		SourceKonto = getKonto(Source);
		if (SourceKonto == 0) return FALSE;
	}// if
  
	// Checken ob TargetKonto existierrt
	TargetKonto=getKonto(Target);
	if (TargetKonto==0) return FALSE;

	// Konten sind existent
	if (Modus==4) SourceKonto->addAktion(*Aktion);
	TargetKonto->addAktion(*Aktion);
	return TRUE;
}

double 
Bank::getKontoStand(int Nummer)
{
	return (getKonto(Nummer))->getKontoStand();
}


TList<TTransaktion>* 
Bank::getAktionen(int Nummer)
{
	return (getKonto(Nummer))->getAktionen();
}

TKunde* 
Bank::getKunde(int Nummer)
{
	for (TIterator<TKunde> i(*KundenListe); i.actualElement()!=0; i++)
		if ( (i.actual()).getNummer()==Nummer ) 
			return &i.actual();

	return NULL;
}

TKunde* 
Bank::getKunde(const TString &Name)
{
	for (TIterator<TKunde> i(*KundenListe); i.actualElement()!=0; i++)
		if (  (i.actual()).getName() == Name ) 
			return &i.actual();  
	return NULL;
}

TKonto* 
Bank::getKonto(int Nummer)
{
	for (TIterator<TKonto> i(*KontoListe); i.actualElement()!=0; i++)
		if ( (i.actual()).getKontoNr()==Nummer ) 
			return &i.actual();

	return NULL;
}

bool 
Bank::deleteKonto(int Nummer)
{
	TKonto *buffer;
	if ((buffer = getKonto(Nummer)) != 0)
		return KontoListe->del(*buffer);
	else 
		return FALSE;
}


bool 
Bank::isAuthorized(int nummer, const TString &name,
				   const TString &pass,int login = 0)
{ 
	if (getKunde(nummer) != 0)
    {
		if ( login == 0 )
			// login == 0 Kunde ist bereits eingeloggt
			// Eines der Passworter und Nummer muessen stimmen
			return   ( pass=="RSA" ||
					   ((getKunde(nummer))->getPw() == pass) ||
					   ((getKunde(nummer))->getSessionPw() == pass)) ;
		else if ( login==1 )
			// login==1 Kunde will mit normalem Passwort einloggen 
			// Passwort und Nummer muessen stimmen
			return (((getKunde(nummer))->getPw())==pass);
		else if ( login ==2 )
			// login==2 Kunde will mit OneTime-Passwort einloggen 
			// OneTimePasswort und Nummer muessen stimmen
		{	  
			cerr << (getKunde(nummer))->getOnePw() << endl;
			cerr << pass << endl;
			return (((getKunde(nummer))->getOnePw())==pass);
		}
		return FALSE;
    }		 
	else 
		return FALSE;
}


//
// handles client-request
//
void 
Bank::handleClientRequest(pid_t pid, MSGData &data)
{  
	TKunde *K;

	switch (data.getOperation())
    {      

        case RSA :
            cerr << "TBank::handleClientRequest (...) : rsa msg" << endl;
            //cerr << data.getKundenNr() << endl;
            /*      if (data.getKundenNr()==rsaKdnr)
                    { // Kunde ist autorisiert
                    rsaLogin=true;
                    rsaKdnr=0;
                    sendString(pid,"OK");
                    }	
                    else*/
            { 	   
                K = getKunde(data.getKundenNr());	    
                if (K!=NULL)	      
                    sendString(pid,K->getPubKey());	      	     
                else
                    sendString(pid,"FAILED");	
            }
            break;
      
        case ENDE :
            cerr << "TBank::handleClientRequest (...) :  ende msg" << endl;

#ifdef STREAMIO
            cerr << "Savesuccess: " << save("bankdata.dat") << endl;      
#endif	 
            sendString(pid,"OK Client Disconnected");      
            break;	  
      
        case LOGIN :
            cerr << "TBank::handleClientRequest (...) :  Login msg" << endl;
            K = getKunde(data.getKundenNr());
            if (strlen(data.getKundenPwd()))
            {		      
                if (isAuthorized(data.getKundenNr(),
                                 data.getKundenname(),
                                 encrypt(data.getKundenPwd()),
                                 1))
                {
                    sendString(pid,"OK BANK: Logindata Correct");
                    K->mutateSessionPw(data.getKundenPwd());
                }// if
                else 
                    sendString(pid,"FAILED BANK: Logindata incorrect");
            }
            else 
                sendString(pid,"FAILED BANK: Passwort missing");
            break;

        case ONELOGIN : // OneTimePasswort
            cerr << "TBank::handleClientRequest (...) :  oneLogin msg" << endl;
            if ((K=getKunde(data.getKundenNr()))!=0)
            {			
                K = getKunde(data.getKundenNr());
                K->mutateSessionPw("");	 			
                if ((TBase64::decode(data.getKundenPwd())).length() > 0)
                { 
                    // Dekodierung erfolgreich	     			       
                    if (isAuthorized(data.getKundenNr(),
                                     "",
                                     encrypt(TBase64::decode(data.getKundenPwd())),
                                     2))
                    {
                        // Passwort ist korrekt
                        sendString( pid, "OK BANK: Logindata correct (onetimepw changed)");
                        // Setze Sessionpasswort, Aendere OnePw
                        K = getKunde(data.getKundenNr());
                        K->mutateSessionPw(data.getKundenPwd());
                        K->mutateOnePw(data.getKundenPwd());	
                    }// if
                    else // Passwort ist nicht korrekt
                    {
                        K->mutateSessionPw("");
                        sendString( pid, "FAIL BANK: Logindata incorrect");
                    }// if
                }// if
                else
                {
                    cerr << "TBank::handleClienRequest (...) : password is not in base64" << endl;
                    sendString( pid, "FAIL BANK : given password has not valid base64-format");
                }// else			
            }// if
            else 
                sendString(pid,"FAIL BANK : incorrect Kundennr.");
            break;        
	  
        case KAUSZ :	    
            cerr << "TBank::handleClientRequest (...) :  Kausz. msg" << endl;
            if (isAuthorized(data.getKundenNr(),data.getKundenname(),data.getKundenPwd()))
                if ( getKonto(data.getKontoNr()) != 0 )
                    if ( ((getKonto(data.getKontoNr()))->getPermission(data.getKundenNr())).isIn(TPermission(OWN)) ||
                         ((getKonto(data.getKontoNr()))->getPermission(data.getKundenNr())).isIn(TPermission(READ)) )
                    {				  										
                        TString buffer;
                        buffer="RESULT:\nTransaktionen:";
                        for (TIterator<TTransaktion> i(*((getKonto(data.getKundenNr()))->getAktionen())); i.actualElement() != 0 ;i++)
                        {	      	
                            buffer+="\n";
                            buffer+=(i.actual()).toString();
                        }// for					
                        buffer+="\n.";
                        sendString(pid,buffer,0);					
                    }      
                    else
                        sendString( pid, "FAIL BANK: No Permission");
                else 
                    sendString( pid, "FAIL BANK: That Konto is not available"); 
            else 
                sendString( pid, "FAIL BANK: Logindata incorrect"); 
            break;

        case KSTAND :
#ifdef DEBUG
            cerr << "TBank::handleClientRequest (...) :  Kstand msg" << endl;
#endif	 	        
            if (isAuthorized(data.getKundenNr(),data.getKundenname(),data.getKundenPwd()))
                if ( getKonto(data.getKontoNr()) != 0 )
                    if ( ((getKonto(data.getKontoNr()))->getPermission(data.getKundenNr())).isIn(TPermission(OWN)) ||
                         ((getKonto(data.getKontoNr()))->getPermission(data.getKundenNr())).isIn(TPermission(READ)) )
                    {
                        if (getKonto(data.getKontoNr())!=0)
                            sendString(pid, getKontoStand(data.getKontoNr()));
                        else 
                            sendString(pid,"FAIL BANK: That Konto is not available");
                    }// if
                    else 
                        sendString(pid,"FAIL BANK: No Permission");
                else 
                    sendString(pid,"FAIL BANK: That Konto is not available"); 
            else 
                sendString(pid,"FAIL BANK: Logindata incorrect");
            break;

        case UEBERW :        	
#ifdef DEBUG
            cerr << "TBank::handleClientRequest (...) :  ueberw. msg" << endl;
#endif
            if (isAuthorized(data.getKundenNr(),data.getKundenname(),data.getKundenPwd()))
                if ( getKonto(data.getKontoNr()) != 0 )
                    if ( getKonto(data.getZielKontoNr()) != 0 )
                        if ( ((getKonto(data.getKontoNr()))->getPermission(data.getKundenNr())).isIn(TPermission(OWN)) ||
                             ((getKonto(data.getKontoNr()))->getPermission(data.getKundenNr())).isIn(TPermission(WRITE)))	      
                        {	  
                            if (addTransaktion(data.getKontoNr(),
                                               data.getZielKontoNr(),
                                               data.getBetrag(),
                                               data.getZweck(),
                                               Ueber))
                                sendString(pid,"OK BANK: Tansaction done");
                            else 
                                sendString(pid,"FAIL BANK: Error doing transaction");		
                        }// if
                        else 
                            sendString(pid,"FAIL BANK: NO Permission for Source Konto");  
                    else 
                        sendString(pid,"FAIL BANK: Target Konto not available"); 
                else 
                    sendString(pid,"FAIL BANK: That Konto is not available"); 
            else 
                sendString(pid,"FAIL BANK: Logindata incorrect");	
            break;
        case RECHTE :
#ifdef DEBUG
            cerr << "TBank::handleClientRequest (...) :  Rechte msg" << endl;
#endif
            if (isAuthorized(data.getKundenNr(),data.getKundenname(),data.getKundenPwd()))
                if ( getKonto(data.getKontoNr()) != 0 )
                    if ( getKunde(data.getZielnummer()) != 0 )
                        if ( ((getKonto(data.getKontoNr()))->getPermission(data.getKundenNr())).isIn(TPermission(OWN)) )
                        {
                            // AccessControlListe des Kontos besorgen	       		
                            TAcl *acl=getKonto(data.getKontoNr())->getAcl();
                            if (data.getRechtKontoauszug()==1)
                                acl->mutateAccess(ADD,READ,data.getZielnummer());
                            if (data.getRechtKontoauszug()==-1)
                                acl->mutateAccess(DEL,READ,data.getZielnummer());
                            if (data.getRechtAbheben()==1)
                                acl->mutateAccess(ADD,WRITE,data.getZielnummer());
                            if (data.getRechtAbheben()==-1)
                                acl->mutateAccess(DEL,WRITE,data.getZielnummer());
                            sendString(pid,"OK BANK: Permission changed");
#ifdef DEBUG
                            cerr  << "permission changed to: " << *acl << endl;
#endif
                        }
                        else 
                            sendString(pid,"FAIL BANK:  Only the Owner can change Permissions");
                    else 
                        sendString(pid,"FAIL BANK: Targetclient not found");
                else 
                    sendString(pid,"FAIL BANK: That Konto is not available");
            else 
                sendString(pid,"FAIL BANK: Logindata incorrect");
            break;
		
        case  SPENDCOIN :
            // 
            // check if the coin is already spended      
            //            
            sendString(pid,spendCoin(data.getSerialNo(),data.getCoinIDs()));     
            break;

        case CHECKSERIAL :
            //
            // look if the given serialno is in coindb
            //
            if (hasSerialNo( data.getSerialNo() ))
            {
                cout << "TBank::handleClientRequest (...) : serialno present" << endl;
                sendString( pid, "TRUE" );
            }// if
            else
            {
                cout << "TBank::handleClientRequest (...) : serialno not present" << endl;
                sendString( pid, "FALSE" );
            }// else
            break;
            
        case NOOP :
        default : // Nothing to do
            break;
    }
}

//
// sends a string to the commu-server with correct pid
void 
Bank::sendString(pid_t pid, const TString &str,int more=0)
{
	struct MessageBuffer mb;   
	str.copyTo( mb.msq_text, sizeof(mb.msq_text) );
	mb.msq_text[str.length()]=0;
	// set message type to pid
	mb.msq_type=pid; 
	if (msgsnd( msqid, (struct msgbuf*) &mb, str.length(), 0) < 0)
		perror ( "Bank::run () : sendMessageError" );

//     cout << "TBank::sendString (" << pid << ",...) : length = " << strlen(mb.msq_text) << endl;
    cout << "TBank::sendString (" << pid << ",...) :    msg = " << mb.msq_text << endl;
  
	if ( more == 0)
    {    
		// END OF TRANSMISSION      
		mb.msq_text[0]=0;      
		if ( msgsnd( msqid, (struct msgbuf*) &mb, 0, 0) < 0 )
			perror ( "Bank::run () : sendMessageError" );  
    }


	  
}

void 
Bank::sendString(pid_t pid, double value)
{
	struct MessageBuffer mb;   
	sprintf(mb.msq_text,"%f",value);
	// set message type to pid
	mb.msq_type=pid;
	if (msgsnd( msqid, (struct msgbuf*) &mb, strlen(mb.msq_text), 0) < 0)
		perror ( "Bank::run () : sendMessage" );    

//     cout << "TBank::sendString (" << pid << ",...) : length = " << strlen(mb.msq_text) << endl;
    cout << "TBank::sendString (" << pid << ",...) :    msg = " << mb.msq_text << endl;

	// END OF TRANSMISSION  
	mb.msq_text[0]=0;
	if (msgsnd( msqid, (struct msgbuf*) &mb, 0, 0) < 0)
		perror ( "Bank::run () : sendMessage" );    
}


//
// handles kommu with kommuservers
//
int
Bank::run ()
{
	struct MessageBuffer mb;
	int      n;
	TMessage msg;
	MSGData  msgData;

	// buf contains information about the message-queue  
	struct msqid_ds buf; 
	
	// recieve all msg with type==1
	while ((n = msgrcv( msqid, (struct msgbuf*) &mb, MSQSIZE, 1, 0)) > 0)
	{
	    mb.msq_text[n]=0;
		// get pid from commu-server
	    msgctl(msqid, IPC_STAT, &buf); 

	    // extract data from string
	    msgData = msg.getData(mb.msq_text);

//         cout << "TBank::run () : received msglength = " << strlen(mb.msq_text) << endl;        
//         cout << "TBank::run () :   received msgfrom = " << buf.msg_lspid << endl;
//         cout << "TBank::run () :       received msg = " << mb.msq_text << endl;

	    // handle request
	    handleClientRequest(buf.msg_lspid,msgData);
	    
	}// while

	return (SUCCESS);
}

//
// output
//
ostream& 
operator << ( ostream& os, Bank &K )
{
	os  << "Kunden: " << endl
		<< *K.KundenListe
		<< "Konten: " << endl
		<< *K.KontoListe ;	  
    return os;
}


//
// saves all bank-data to file
//
bool 
Bank::save(const TString &file)
{
	char f[file.length()+1];
	file.copyTo(f);f[file.length()]=0;
	ofstream out(f,ios::out); // Creates new file
	if ( out )
    { //Append data to file
		out << *this;
		return (true);
    }
	else return (false);  
}


//
// loads all bank-data from file
//
bool
Bank::load(const TString &file)
{
	char f[file.length()+1];
	file.copyTo(f);f[file.length()]=0;
	int Zeilenlaenge=1024; // Maximale Zeilenlaenge betraegt 1024
	char puffer [ Zeilenlaenge ];
	ifstream in(f,ios::in);
	if (in)
    { // Datei ist da
     
		while (in.getline(puffer,Zeilenlaenge))
		{ //Zeilenweises Einlesen der Datei
			cout << puffer << endl;
			if (! strcmp(puffer,"KUNDE"))
			{	   
				cout << "Kunde detected" << endl;
				TKunde Kunde(in);                // Laedt Kundendaten vom Strea
				KundenListe->addElement(Kunde);
			}
			if (! strcmp(puffer,"KONTO"))
			{	   
				cout << "Konto detected" << endl;
				TKonto Konto(in);                // Laedt Kontendaten vom Stream
				KontoListe->addElement(Konto);
			}
		}            
		save("tmp");
		return true;
      
    }
	return false;
}

//
// sign and return given coin
//
TString
Bank::signCoin ( const TString & string )
{
    return "OK";
}

//
// access coin DB
//
bool
Bank::hasSerialNo ( long serialNo ) const
{
    uint  i;
    
    for (i = 0; i < arrCoins.getSize(); i++)
    {
        if (arrCoins[i].getSerialNo() == serialNo)
            return TRUE;
    }// for

    return FALSE;
}

// check the coin if its already spended and returns a message
TString 
Bank::spendCoin( int Sn, const TString & string )
{  
    TString result, subStr;
    TString temp(string);
    TBitArray arrID[NO_OF_ID];
    TCoinData Coin;
    Coin.setSerialNo(Sn);

    subStr=temp.strtok( '[' );
    for (int i = 0; i < NO_OF_ID; i++)
    {	  
        arrID[i] = TBase64::decode( temp.strtok( ',' ) );
        //cout << "Bank::spendCoin (...) : SN:" << Sn << "ID[" << i << "] = " << arrID[i] << endl;
        Coin.arrID[i]=arrID[i];
    }
  
    for (uint i=0; i< arrCoins.getSize(); i++)
        if (Sn==arrCoins[i].getSerialNo())
        {
            //The Coin is already spended	
            for (int j=0; j < NO_OF_ID ; j++)
                if (arrCoins[i].arrID[j] != arrID[j])
                {	     	     	     
                    // We determine the owner of the coin	      
                    TBitArray tmp;
                    tmp=arrID[j];
                    tmp^=arrCoins[i].arrID[j];
                    cerr << "Bank::spendCoin (...) :ID:" << j << " CHEATER ALERT - Coin owner is: " << tmp.toString() << endl;
                    result="C-ERROR: Coin is already spended by owner"; 
                    return  result;
                }	  	    
            cerr << "Bank::spendCoin (...) : all ID's are identical - We couldn't determine the owner of the coin" <<  endl;
            result="S-ERROR: Coin is probably spended twice by store"; 
            return result;	
        }  
    // A new Coin
    // We mark the coin as spended
    arrCoins.setSize(arrCoins.getSize()+1);  
    arrCoins[arrCoins.getSize()-1]=Coin; 
    cerr << "Bank::spendCoin (...) : Coin is accepted and registered" <<  endl;
    result="OK Coin is accepted"; 
    return result;
}
