#ifndef KUNDE_HH
#define KUNDE_HH

#include <iostream.h>
#include <fstream.h>
#include "../misc/TString.hh"

//
// Allgemeine Klase fuer einen Kunden
// Ein Kunde besteht aus seinem Namen, 
// einer Kundennummer und einem Passwort
//


class TKunde
{

protected:
    // Initialisiert das Object
	void init(int nummer, 
			  const TString &name,
			  const TString &passwort,
			  const TString &opassw,
			  const TString &spw,
			  const TString &key );
	TString Name;                                          // Name des Kunden
	int Nummer;                                            // Nummer des Kunden
	TString Passwort;                                      // Passwort des Kunden
	TString SessionPw;                                     // In dieser Session gltiges onetimepasswd
	TString OnePw;                                         // One-Time-Password
	TString pubkey;                                        // RSA-Public-Key

public:
	TKunde();                                                  // Standardkonstruktor
	TKunde(int nummer, 
		   const TString &name,
		   const TString &passwort,
		   const TString &onepw,
		   const TString &pubkey);                              // Legt einen Kunden an
	TKunde(const TKunde &Kunde);                               // Copy-Konstruktor
	TKunde(ifstream&);                                         // Laedt Kundendaten aus Stream
	~TKunde();                                                 // Destruktor
	TKunde& TKunde:: operator = ( const TKunde &TKunde);      // Zuweisungsoperator

	friend bool operator ==( TKunde &K, TKunde &k ) {           // Test auf Gleichheit
		// Zwei Kunden sind gleich wenn Nummer gleich
		return (k.Nummer==K.Nummer); 
	};

	friend ostream& operator <<( ostream& os, TKunde &Kunde )
	{ os << "KUNDE"        << endl
		 << Kunde.Nummer   << endl
		 << Kunde.Name     << endl 
		 << Kunde.Passwort << endl
		 << Kunde.OnePw    << endl 
		 << Kunde.pubkey   << endl;
	//<< Kunde.SessionPw ;  
	return os;
	}
 
 
	TString getName() const;                                      // liefert Namen des Kunde
	TString getPw() const;                                        // liefert Passwort des Kunde
	TString getOnePw() const;                                     // liefert OneTimePasswort
	TString getSessionPw() const;                                 // Liefert SessionPasswort
	TString getPubKey() const;                                    // liefert RSA-Public-Key des Kunden
	int getNummer() const;                                         // liefert Nummer des Kunde  

	// Aendert Namen des Kunden
	void mutateName(const TString &name);

	// Aendert Nummer des Kunden
	void mutateNummer(int Nummer);

	// Aendert Passwort des Kunden
	void mutatePw(const TString &passwort);

	// Aendert SessionPasswort des Kunden
	void mutateSessionPw(const TString &passwort);

	// Aendert OneTimePasswort des Kunden
	void mutateOnePw(const TString &passwort);
};
  

#endif // KUNDE_HH








