#ifndef __TBITARRAY_HH
#define __TBITARRAY_HH
//
//  Project  : libmisc
//  File     : TBitArray.hh
//  Author   : Ronald Kriemann
//  Purpose  : class for array holding bits
//

#include <iostream.h>

#include "defines.hh"
#include "dynarray.hh"
#include "TString.hh"

// used type to store bits
typedef ulong TStoreType;

//
//
//
class TBitArray : public TDynArray<TStoreType>
{
protected:
	// number of stored bits
	ulong noOfBits;

	// size of the used type to store bits
	static ulong sizeOfBase;

public:
	// constructor
	TBitArray ( ulong n = 0 );

	// get size and resize
	virtual const uint getSize () const;
	virtual void setSize ( uint n );

	// access bits (not possible via operator[])
	const bool getBit ( uint n ) const;
	void setBit ( uint n, bool b );

	// init bitarray with random bitstream
	void initRandom ( const ulong &seed = 0 );

	// XOR operator
	TBitArray & operator ^= ( const TBitArray &a );

	// copy operator
	TBitArray & operator = ( const TBitArray &a );
	TBitArray & operator = ( const TString &string );

	// compare
	const bool operator == ( const TBitArray &a ) const;
	const bool operator != ( const TBitArray &a ) const;

	// export to string
	TString toString () const;

	// output
	friend ostream & operator << ( ostream &os, const TBitArray &a );
};

#endif  // __TBITARRAY_HH
