#if !defined(__TCOIN_HH)
#define __TCOIN_HH

//
//  Project   : libmisc
//  File      : TCoin.hh
//  Author    : Ronald Kriemann
//  Purpose   : class for a digital coin
//

#include "defines.hh"
#include "TString.hh"
#include "TBitArray.hh"

// some sizes
#define  ID_SIZE         256
#define  BIT_COMMIT_SIZE 256
#define  NO_OF_ID        8
#define  NO_OF_COIN      8

//
//
//
class TCoin
{
protected:
	// personal id
	TBitArray  id;

	// value of this particular 
	TFloat   value;

    // serial no.
    long serialNo;

    // the coin as string
    TString  szCoin;
    
	// signed representation of this coin
	TString  signedCoin;

    // bit-strings for id-check
    TBitArray  arrID[NO_OF_ID][2];
    
public:
	// constructor
	TCoin ( const TFloat v, const TString &strID );
	TCoin ( const TString &s );

	// prepare unsigned coin for signing by bank
	TString prepare ( const TString &bankID );

    // access i'th ID-bitstring (index 0 or 1)
    const TBitArray & getBlindFac ( const uint &i, const uint &j ) const;

    // access signed coin
    const TString & getSignedCoin () const;
    TCoin & setSignedCoin ( const TString & string );

    // access coin string
    const TString & getCoinStr () const;

    // copy operator
    TCoin & operator = ( const TCoin & );
    
protected:
    // build next serial no.
    long buildNextSerialNo ();
};

#endif  // __TCOIN_HH
