//
//  Project  : libmisc
//  File     : TException.cc
//  Author   : Ronald Kriemann
//  Purpose  : baseclass for all exceptions
//

#include "TException.hh"

//
// constructor
//
TException::TException ()
{
}

//
// destructor
//
TException::~TException ()
{
}

//
// output
//
void
TException::print ( ostream & os ) const
{
    os << "TException";
}

//
// output (operator)
//
ostream &
operator << ( ostream & os, const TException & exception )
{
    exception.print( os );

    return os;
}

/////////////////////////////////////////////////////
//
// exception for range-overflow
//

//
// constructor
//
TRange::TRange ()
{
    pos  = max  = 0;
    bPos = bMax = FALSE;
}

TRange::TRange ( size_t p )
{
    ::TRange();

    pos  = p;
    bPos = TRUE;
}

TRange::TRange ( size_t p, size_t m )
{
    ::TRange();

    pos  = p;
    bPos = TRUE;

    max  = m;
    bMax = TRUE;
}

//
// access pos and bound
//
const size_t
TRange::getPos () const
{
    return pos;
}

const size_t
TRange::getMax () const
{
    return max;
}

//
// output
//
void
TRange::print ( ostream & os ) const
{
    os << "TRange";

    if (bPos && !bMax)
        os << " (pos = " << pos << ")";

    if (bPos && bMax)
        os << " (pos = " << pos << ", max = " << max << ")";
}

/////////////////////////////////////////////////////
//
// exception for wring sizes
//

//
// constructor
//
TSize::TSize ()
{
    size  = 0;
    bSize = FALSE;
}

TSize::TSize ( size_t s )
{
    size  = s;
    bSize = TRUE;
}

//
// access size
//
const size_t
TSize::getSize () const
{
    return size;
}

//
// output
//
void
TSize::print ( ostream & os ) const
{
    os << "TSize";

    if (bSize)
        os << "(size = " << size << ")";
}

/////////////////////////////////////////////////////
//
// baseclass for mathematical exceptions
//

//
// output
//
void
TMathErr::print ( ostream & os ) const
{
    os << "TMathErr";
}

/////////////////////////////////////////////////////
//
// class for division by zero exception
//

//
// output
//
void
TDivByZero::print ( ostream & os ) const
{
    os << "TDivByZero" << endl;
}

/////////////////////////////////////////////////////
//
// class for arithmetic underflow exceptions
//

//
// output
//
void
TUnderFlow::print ( ostream & os ) const
{
    os << "TUnderFlow" << endl;
}
