//
//  Project   : libmisc
//  File      : TServerSocket.cc
//  Author    : Ronald Kriemann
//  Purpose   : class for socket-based communication (server side)
//

extern "C" {
#include <stdio.h>
#include <unistd.h>
#include <string.h>

#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
}

#include "TServerSocket.hh"

//
// constructor
//
TServerSocket::TServerSocket ( uint port )
{
	setPort( port );
	initSocket();

	memset( &name, 0, sizeof( struct sockaddr_in ));
}

//
// bind to socket
//
int 
TServerSocket::bind ()
{
	int  n, len;

	name.sin_family = AF_INET;
	name.sin_port   = htons(getPort());

	len = sizeof( struct sockaddr_in );
	n   = INADDR_ANY;
	memcpy(&name.sin_addr, &n, sizeof(long));

	// bind the socket to the adress
	if (::bind(sd, (struct sockaddr *) &name, len) < 0)
	{
		perror( "TServerSocket::bind () : bind" );
		return ERROR;
	}// if

	return SUCCESS;
}

//
// listen on socket
//
int 
TServerSocket::listen ()
{
	if (::listen(sd, 5) < 0)
	{
		perror( "TServerSocket::listen () : listen" );
		return ERROR;
	}// if

	return SUCCESS;
}

//
// wait for client to connect and return new socket descr.
//
int 
TServerSocket::accept ()
{
	int  len, ns;

	if ((ns = ::accept(sd, (struct sockaddr *) &name, &len)) < 0)
	{
		perror( "TServerSocket::accept () : accept" );
		return ERROR;
	}// if

	return ns;
}

//
// start whole communication
//
int 
TServerSocket::startCommu ()
{
	if (bind() != SUCCESS)
		return ERROR;

	if (listen() != SUCCESS)
		return ERROR;

	return accept();
}
