//
//  Project   : libmisc
//  File      : base64.hh
//  Author    : Christian Rost / Ronald Kriemann
//  Purpose   : class for base64 encoding and decoding
//

#include <string.h>
#include <iostream.h>

#include "base64.hh"

//
// available char.
//
char    TBase64::base64CharArr[] = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/=";
TString TBase64::base64CharStr   = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/=";

//
// constructor
//
TBase64::TBase64()
{
}

//
// encode bytestream into base64
//
void 
TBase64::encode( uchar output[], const uchar input[], int len )
{
	int    c, n, i, ii, ci;
	long   val;
	char   enc[4];

	// buffer for encoding
	uchar  buffer[len*4];

	n = val = ii = 0;

	//
	// encode main part
	//	
	for ( ci = 0 ; ci < len ; ci++) 
	{
		c = input[ci];
		if (n++ <= 2) 
		{ 
			val <<= 8;
			val  += c;      
			continue;
		}// if
      
		for (i = 0; i < 4; i++) 
		{
			enc[i]   = val & 63;
			val    >>= 6;
		}// for
      
		for (i = 3; i >= 0; i--)      
			buffer[ii++] = base64CharArr[enc[i]];

		n   = 1;
		val = c;
	}// for

	// 
	// care about the rest
	//

	if (n == 1) 
	{
		val <<= 16;
		for (i = 0; i < 4; i++) 
		{
			enc[i]   = val & 63;
			val    >>= 6;
		}// for
		enc[0] = enc[1] = 64;
	}// if

	if (n == 2) 
	{
		val <<= 8;
		for (i = 0; i < 4; i++) 
		{
			enc[i]   = val & 63;
			val    >>= 6;
		}// for
		enc[0] = 64;
	}// if
  
	if (n == 3)
	{
		for (i = 0; i < 4; i++) 
		{
			enc[i]   = val & 63;
			val    >>= 6;
		}// for
	}// if

	if (n) 
	{
		for (i = 3; i >= 0; i--)     
			buffer[ii++] = base64CharArr[enc[i]];    
	}// if

	buffer[ii]=0;
	strcpy( (char*) output, (char*)buffer );
}

void 
TBase64::encode( TString &output, const TString &input )
{
	int    c, n, i, ii, ci;
	long   val;
	char   enc[4];
	long   len;

	len = input.length();

	if (len > 0) {
	  
	
	output.setLength( len*4 );

	n = val = ii = 0;

	//
	// encode main part
	//	
	for ( ci = 0 ; ci < len ; ci++) 
	{
		c = input[ci];
		if (n++ <= 2) 
		{ 
			val <<= 8;
			val  += c;      
			continue;
		}// if
      
		for (i = 0; i < 4; i++) 
		{
			enc[i]   = val & 63;
			val    >>= 6;
		}// for
      
		for (i = 3; i >= 0; i--)      
			output[ii++] = base64CharStr[enc[i]];

		n   = 1;
		val = c;
	}// for

	// 
	// care about the rest
	//

	if (n == 1) 
	{
		val <<= 16;
		for (i = 0; i < 4; i++) 
		{
			enc[i]   = val & 63;
			val    >>= 6;
		}// for
		enc[0] = enc[1] = 64;
	}// if

	if (n == 2) 
	{
		val <<= 8;
		for (i = 0; i < 4; i++) 
		{
			enc[i]   = val & 63;
			val    >>= 6;
		}// for
		enc[0] = 64;
	}// if
  
	if (n == 3)
	{
		for (i = 0; i < 4; i++) 
		{
			enc[i]   = val & 63;
			val    >>= 6;
		}// for
	}// if

	if (n) 
	{
		for (i = 3; i >= 0; i--)     
			output[ii++] = base64CharStr[enc[i]];    
	}// if

	output.setLength( ii );
	}
}

TString
TBase64::encode(const TString &input)
{
	TString result;

	encode( result, input );
	return result;
}

//  
// decode base64 to bytestream
//
int
TBase64::decode( uchar output[], const uchar input[] )
{
	int	   i, num, len, j, ii, iii;
	long   d, val;
	uchar  nw[4], *p, *c;	
 
	iii = 0;

	len = strlen((char*)input);
	for (i = 0; i < len; i += 4)
	{
	    val = 0;
	    num = 3;
	    c   = const_cast<uchar*>(input+i);
	    if (c[2] == '=')
			num = 1;
	    else if (c[3] == '=')
			num = 2;
	    
	    for (j = 0; j <= num; j++)
		{
			if (!(p = (uchar*)strchr(base64CharArr, c[j])))
				// not a base64 - string
				return 0;

			d    = p - (uchar*)base64CharArr;
			d  <<= (3-j)*6;
			val += d;
		}// for

	    for (j = 2; j >= 0; j--)
		{
			nw[j]   = val & 255;
			val   >>= 8;
		}// for
	    
	    for (ii = 0; ii < num; ii++)
			output[iii++] = nw[ii];

	    output[iii] = 0;
	}// for

	return(iii-1);
}

int
TBase64::decode( TString &output, const TString &input )
{
	int	   i, num, len, j, ii, iii;
	long   d, val;
	uchar  nw[4];
	TString p, c;
 
	iii = 0;

	len = input.length();
	output.setLength( len );
	for (i = 0; i < len; i += 4)
	{
	    val = 0;
	    num = 3;
	    c   = input.index( i );
	    if (c[2] == '=')
			num = 1;
	    else if (c[3] == '=')
			num = 2;
	    
	    for (j = 0; j <= num; j++)
		{
			//			if (!(p = static_cast<uchar*>(strchr(base64Char, c[j]))))
			p = base64CharStr.strchr( c[j] );
			if (p.isEmpty()) return 0;

			d    = base64CharStr.length() - p.length();
			d  <<= (3-j)*6;
			val += d;
		}// for

	    for (j = 2; j >= 0; j--)
		{
			nw[j]   = val & 255;
			val   >>= 8;
		}// for
	    
	    for (ii = 0; ii < num; ii++)
			output[iii++] = nw[ii];
	}// for

	output.setLength( iii );
	return iii;
}

TString
TBase64::decode(const TString &input)
{
	TString result;

	if (!decode( result, input ))
		result.setLength( 0 );

	return result;
}
