#ifndef __MD5_HH
#define __MD5_HH
//
//	Project		: besy
//	File		: md5.hh
//	Author		: Ronald Kriemann
//	Purpose		: class for the MD5 algorithm (see RFC 1321)
//	Date		: 16.05.1997
//	Changes     : 
//

#include "defines.hh"
#include "TString.hh"

//
//
//
class TMD5 
{
protected:
	// state (ABCD)
	UINT4  state[4];
	// number of bits, modulo 2^64 (lsb first)
	UINT4  count[2];
	// input buffer
	uchar  buffer[64];

public:
	// constructor
	TMD5 ();
	// destructor
	~TMD5 ();

	// crypt. given input into digest
	void crypt ( uchar digest[16], const uchar input[], uint len );
	TString crypt( const TString &input );

	// MD5 block update op., continues an MD5 msg-digest op, processing
	// another msgblock and updating the context (state, count etc)
	void update ( uchar input[], uint len );

	// finish the MD5-algo
	void finish ( uchar digest[16] );

	// really do the MD5-Algo
	void transform ( uchar block[64] );

	// encode input (UINT4) into output (uchar). assume len is a mult. of 4
	void encode ( uchar output[], UINT4 input[], uint len );

	// decode input (uchar) into output (UINT4). assume len is a mult. of 4
	void decode ( UINT4 output[], uchar input[], uint len );
};

#endif  // __MD5_HH
