//
//  Project   : besy
//  File      : store.cc
//  Author    : Christian Rost
//  Purpose   : class for a store


extern "C"
{
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <signal.h>
#include <string.h>

#include <sys/wait.h>
#include <sys/ipc.h>
#include <sys/msg.h>
}

#include <iostream.h>

#include "../misc/base64.hh"
//#include "../misc/TRandom.hh"

#include "../misc/msqdef.hh"
#include "../misc/scanner.hh"
#include "../misc/md5.hh"
#include "../misc/TRandom.hh"

#include "store.hh"

// indicates if we have to finish
bool  bFinishWork;


// defines for rsa import
#define BANK    1
#define PUBLIC  1

// some defines for transmitting a coin
#define NOTHING           0
#define STARTTRANSMITCOIN 1
#define TRANSMITCOIN      2
#define TRANSMITID        3 

//
// signalhandler for SIGCHILD
//
// just gets the exitstatus and returns
//
void
sigHandler (int signr)
{
#ifdef DEBUG
	psignal( signr, "sigHandler (int) : recieved signal" );
#endif
	// handle signal
	switch (signr)
	{
        case SIGCHLD : // recieve childstatus
            waitpid( 0, (int*) NULL, WNOHANG );
            break;
        case SIGTERM : // finish process
            //	case SIGINT  :
            bFinishWork = TRUE;
            break;
        default      :
            break;
	}// switch
}

//
// constructor
//
TStore::TStore ( int portnumber )
{
	Socket.setPort(portnumber);
	bForked     = FALSE;
	msqid       = -1;
	bFinishWork = FALSE;
	init();
}

//
// destructor
//
TStore::~TStore ()
{
}

//
// create the serversocket / and install signalhandler
//
void 
TStore::init ()
{
    struct sigaction rTERMAction;
  
    if (Socket.create() != SUCCESS)
    {
        cerr << "Server::init () : unable to create socket" << endl;
        exit( -1 );
    }// if
  
    // install all needed signalhandler
    signal(SIGCHLD, SIG_IGN);
  
    rTERMAction.sa_handler = (void (*) (int)) sigHandler;
    sigemptyset( &(rTERMAction.sa_mask) );
    sigaddset(   &(rTERMAction.sa_mask), SIGTERM );
    rTERMAction.sa_flags   = 0;
    sigaction( SIGTERM, &rTERMAction, NULL);
    transmitCoinState=NOTHING;    

    rsaBank.setVerbosity( 1 );
  
    // load all keys  
    loadKey( BANK, 14, PUBLIC );

}

//
// wait for clients
//
int 
TStore::wait ()
{  
    int   sd;
  
    // wait for clients
    while (bFinishWork == FALSE)
    {            
        // check if clients have connected
        if ((sd = Socket.select()) >= 0)
        {
            // a client has connected and we have a communicationsocketdescriptor (nsd)  
            cout << "TStore::wait () : client has connected" << endl;

            // fork another server
            pid_t pid = fork();
	  
            // check if forked
            // then the parent starts commu with client and 
            // the child is listenning on the socketServer
            if (pid < 0)
            {
                perror( "TStore::wait () : ERROR while forking" );
                exit( ERROR );
            }// if
            else if (pid == 0)
            {
	  
                cout << "TStore::wait () : coummnication-child has forked" << endl;
	  
                bForked = TRUE; //I'm a child now
				// get the msqid for bank-communication
                key_t  key = MSQKEY;
                if ((msqid = msgget(key, MSQPERM)) < 0)
                {
                    perror( "TStore::wait () : ERROR while msgget" );
                    return (ERROR);
                }// if
				// start commu with clients through nsd
	  
                cout << "TStore::run () : starting commu with client" << endl;
                sendln( sd, "Hello. You are connected to Amnesty International. Send a coin please :)");	  
                return (run(sd));
            }// else
        }// if
      
    }// while
    return (SUCCESS);
}

//
// do your job
//
int 
TStore::run (int nsd)
{
    MSGData  msgData;                  // Data for communication through msg_queue to bank
    bool     bEndCommu = FALSE;        // SocketCommunication alive?   
    TString  socketmsg;                // Message from socket 
  
    msgData.clear();
    
    while (bEndCommu == FALSE)       
    {
        // Waiting for a line\n from Client
        if (Socket.recvln(nsd,socketmsg) < 0) 
            // Client closes Connection
            bEndCommu = TRUE;      
	
        if (socketmsg.length() > 0)
        {	      	  	  
            // socketmsg enthaelt nun die erhaltene Nachricht
            // remove "\n"
            socketmsg.setLength(socketmsg.length()-1);
	  
            // parse message
            parseMsg( nsd, socketmsg, msgData, bEndCommu );		   
        }//if [Nachricht vorhanden]            
        else
        {
            bEndCommu = TRUE;
        }// else
    }// while  

    cout << "TStore::run () : finished client communication" << endl;

    // close communication socket of this client
    Socket.close(nsd);  
    return SUCCESS;
}

//  
// do parsing of messages
//
bool 
TStore::parseMsg ( int  nsd, TString plainMsg, MSGData &msgData, bool &bEndCommu )
{
    TMessage msg;
    Scanner::TokenData data;
    bool     bSendMsg  = FALSE;
  
    Scanner scanner( plainMsg );
    //cerr << "RECEIVED:" << plainMsg << endl;  
    while ((scanner.nextToken(data) == SUCCESS) && (data.token != tokEOL))
    {	     
        switch (data.token)
        {	 	
            case tokENDE :                               // ENDE	 
                cout << "TStore::parseMsg () : closing connection to client" << endl;	  
                msgData.clear();	  
                msgData.setOperation(ENDE);	  
                msg.buildBankMsg(msgData);	  
                bSendMsg = TRUE; 
                sendln(nsd,"OK Connection closed");	  
                // close connection				 
                bEndCommu = TRUE;	  
                break;	 	 	  
            case tokOP :                                // OPERATION
                cout << "TStore::parseMsg () : Operation requested" << endl;	 
                scanner.nextToken(data);	 	  	  	
                switch (data.token)
                {	     		      			      		
                    case tokOPSignCoin :                  // OP SIGNCOIN (ReceiveCoin)
                        cout << "TStore::parseMsg () :  Receiving a coin" << endl;		
                        receiveCoin( nsd, data.szString );
                        break;
		
                    default :                           // unknown OP
                        cout << "TStore::parseMsg () : unknown operation : Token:" << data.token << " "
                             << data.szString << endl;			
                        transmitCoinState = NOTHING;
                        break;
                }// switch	 
                break;	  
            case tokNothing :	    
            default      :                               // unknown command	
                cout << "TStore::parseMsg () : unknown command : " 
                     << plainMsg << endl;	       
                sendln( nsd, "FAILED Unknown Command");
                break;
        }// switch
    }// while
  
    cout << "TStore::parseMsg () : end of scanning msg, op_req = "<< bSendMsg << endl;
  
    // Operation requested ?
    /* if (bSendMsg)
       {					      		    		    
       // delete old operation-data except logininformation and kontonr.
       msgData.clear();
       bSendMsg = FALSE;
       //send Message to Bank and send Result to clients
       return sendln( nsd, sndMsg(msg) );
       } // end of Operation*/
  
    return SUCCESS;
}




//Send Message to Bank and return the resulting message
TString
TStore::sndMsg ( TMessage &msg )
{  
    int                  size;
    bool                 bEndOfTransmission = FALSE;
    struct MessageBuffer mb;
  
    // for ident. we set the mb.type to our pid
    mb.msq_type = getpid();
  
    // send command, coded with our pid, to bank
    cout << "TStore::sndMsg (TMessage) : length of msg = " << strlen(msg.getMsg()) << endl;
    strcpy(mb.msq_text,msg.getMsg());  

    // Message-Type == Receiver is 1  (==server)
    mb.msq_type=1;

    //
    // send msg to bank
    //
    // 	cout << "Server::sndMsg () : sending message to bank [id = " << msqid << "]:"
    // 		 << mb.msq_text << "[len = " << strlen(mb.msq_text) << "]" << endl;
    if (msgsnd( msqid, (struct msgbuf*) &mb, strlen(msg.getMsg()), 0) < 0)
    {
        perror ( "Server::sndMsg () : error in msgsnd" );
        return "Error while sending Message to bank";
    }// if
  
    //
    // wait for messages from bank with our pid and pass it to client
    //
    // 	cout << "Server::sndMsg () : waiting for message from bank: " << endl;
    while ( !bEndOfTransmission &&
            (size = msgrcv( msqid, (struct msgbuf*) &mb, MSQSIZE, getpid(), 0)) > 0)
    {				  				  
        mb.msq_text[size] = 0;
    }// while
  
    // 	cout << "Server::parseMsg () : message form bank received: " << mb.msq_text << endl;
  
    return mb.msq_text;
}

//
// sends the line + \n to a socket
//
bool 
TStore::sendln ( int nsd, const TString & string )
{
    TString  szMsg; 
    szMsg = string;
    return Socket.sendln( nsd, szMsg );
}



void
TStore::receiveCoin(int clientSocket ,const TString &socketMsg)
{
    TString temp(socketMsg), subStr;
    TString coin;
    int i;
    TFloat   value;   
    int      charSize = TString::charSize();  
    TMD5     md5;

  
    if (transmitCoinState == NOTHING)
        transmitCoinState++;
    if (transmitCoinState == STARTTRANSMITCOIN)
    {
        if (temp=="START")
        {
            cout << "TStore::receiveCoin () : Start Coin Communication:" << endl;             
            sendln( clientSocket, "OK" );
            transmitCoinState++;
        }
        else
        {
            sendln( clientSocket, "ERROR" );
            cout << "TStore::receiveCoin () : Abort Coin Communication"<< temp << endl;             
            transmitCoinState=NOTHING;
        }     
    }
    else
    {
        if (transmitCoinState == TRANSMITCOIN)
        {             
            coin = temp.strtok( '|' );
            //cout << "TStore::receiveCoin () : Coinstring:" << coin << endl;
            temp=socketMsg;
      
            ////////////////////////////////////////////////////////////
            //
            // check coin for correct format and blindfac
            //
            
            // the value
            subStr = temp.strtok( ',' );
            subStr >> value;
            cout << "TStore::receiveCoin (...) :    Coinvalue = " << value << endl;
      
            // the bankID
            subStr = temp.strtok( ',' );
            cout << "TStore::receiveCoin (...) :   CoinbankID = " << subStr << endl;      
      
            // the serial no
            subStr = temp.strtok( ',' );
            subStr >> serialNo;
            cout << "TStore::receiveCoin (...) : CoinSerialNo = " << serialNo << endl;
      

            // the md5'ed blind factors
            for (i = 0; i < NO_OF_ID; i++)
            {
                temp.strtok( '(' );
	  
                subStr  = TBase64::decode( temp.strtok( ',' ) );
                subStr += TBase64::decode( temp.strtok( ')' ) );
	  
                arrBlindMD5[i][0] = subStr.substr( 0, 128 / charSize );
                arrBlindMD5[i][1] = subStr.substr( (128 + BIT_COMMIT_SIZE) / charSize
                                                   , 128 / charSize );
	  
                arrR[i][0] = subStr.substr( 128 / charSize, BIT_COMMIT_SIZE / charSize );
                arrR[i][1] = subStr.substr( (256 + BIT_COMMIT_SIZE) / charSize
                                            , BIT_COMMIT_SIZE / charSize );
	  
	  
                if ((arrBlindMD5[i][0].length() != 16) || (arrBlindMD5[i][1].length() != 16) ||
                    (arrR[i][0].length() != 32) || (arrR[i][1].length() != 32))	    
                    cerr << "TStore::receiveCoin (...) : md5 and blindfac do not have correct length" << endl;	   
	  
                temp.strtok( ',' );	  
            }// for
      
            temp.strtok( '|' );
            //
            // get and check Signature
            //                 
            subStr  = temp.strtok( '.' ) ; 
            TBigInteger <RSA_BASE> bnHash(TBase64::decode(subStr));
            bnHash.powerMod(rsaBank.getPublicKey(),rsaBank.getModulus());
            if (bnHash.toString() != md5.crypt(coin))
            {
                cerr << "TStore::receiveCoin (...) : Wrong coin signature " << endl;	  
                sendln( clientSocket, "ERROR: Wrong coin signature " );    
                transmitCoinState=NOTHING;
            }
            else
            {
                temp  = "OK (";      
                for (i=0; i< NO_OF_ID; i++)
                {
                    char v[3];
                    double rand = TRandom::random();
                    if (rand >= 0.5) randList[i]=1; else randList[i]=0;
                    sprintf(v,"%i,",randList[i]);
                    temp+=v;	  
                }
                temp+=")";      
                sendln( clientSocket, temp );      
                transmitCoinState++;      
            }  
        }  
        else
        {
            if (transmitCoinState == TRANSMITID)
            {                        
                //
                // get and check blindfactors
                //	
                subStr=temp.strtok( '(' );
                for (i = 0; i < NO_OF_ID; i++)
                {	  	     
                    arrID[i] = TBase64::decode( temp.strtok( ',' ) );	  	       
                    //cout << "TStore::receiveCoin (...) : ID[" << i << "] = " << arrID[i] << endl; 	  	  	  
                }// for
	
                int incorrect=0;
                for (i = 0; i <  NO_OF_ID; i++)
                {	  
                    temp=arrR[i][randList[i]];
                    temp+=arrID[i].toString();
                    if (arrBlindMD5[i][randList[i]]==md5.crypt(temp))	    	    	    
                        cerr << "TStore::receiveCoin (...) ID[" << i << "] correct" << endl;	      	    
                    else
                    {
                        cerr << "TStore::receiveCoin (...) ID[" << i << "] incorrect" << endl;	    	  
                        incorrect++;	      
                    }
                }// for
                if (!incorrect)
                {
                    TMessage msg;
                    MSGData  msgData;
                    TString tmp="[";
                    // Sending Coin to Bank
                    cerr << "TStore::receiveCoin (...) Send Coin to Bank" << endl;
                    msgData.setOperation( SPENDCOIN );
                    msgData.setSerialNo(serialNo);
                    for (i = 0; i <  NO_OF_ID; i++)
                    {
                        tmp+=TBase64::encode(arrID[i].toString());
                        tmp+=",";
                    }
                    tmp+="]";
                    msgData.setCoinIDs(tmp);	    
                    msg.buildBankMsg( msgData );
                    tmp=sndMsg(msg);
                    cerr << "TStore::receiveCoin (...) Bank says: " << tmp << endl;
                    if (tmp[0]=='O')
                        sendln( clientSocket, "Bank accepts your coin" );
                    else if (tmp[0]=='C')
                        sendln( clientSocket, "Cheater ! Bank doesn`t accept your coin" );
                    else if (tmp[0]=='S')
                        cerr << "TStore::receiveCoin (...) Ups. We probably send a coin two times"
                             << endl;
                }// if
                transmitCoinState=NOTHING;
            }// if
        }// else
    }// else
}



bool
TStore::loadKey ( int owner, int number, int type )
{
    FILE  *file;
    char   buffer[16384];
    int    n;
    char   key[16384];

    switch (type)
    {
        case PUBLIC  : file = fopen( "../keys/public_keys", "r" ); break;
        default        : return FALSE;
    }// switch
    
    if (file == NULL)
    {
        cerr << "Server::loadKey (...) : could not open file" << endl;
        return FALSE;
    }// if
    
    while (!feof( file ))
    {
        fgets( buffer, sizeof(buffer), file );
        
        memset( key, 0, sizeof(key) );
        n = -1;
        
        switch (owner)	  
        {
            case BANK   : sscanf( buffer, "B_BESY%d,%s", &n, (char*) &key ); break;	    
            default       :
                fclose( file );
                return FALSE;
        }
     
        if (n == number)
        {                                 
            switch (owner)
            {
                case BANK   : rsaBank.setModulus( TBase64::decode( key ) ); break;	    
            }// switch      
	    
            fclose( file );	
            return TRUE;
        }// if
    }// while
    
    fclose( file );
    
    return FALSE;
}



